/*---------------------------------------------------------------------
 * nano_os.h
 *
 * Copyright (c) 2010 Tobias Wiese - hacker-tobi@gmx.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *------------------------------------------------------------------------
 */


//+++++++++++++++++++++GLOBAL DEFINE+++++++++++++++++++

// undefine to save memory if signal & wait are not needed
// #define TASK_USE_SLEEP

// undefine to save memory if task anhilate is not needed
#define TASK_USE_ANHILATE

// undefine to save memory if message queue is not needed
#define TASK_USE_MESSAGE

// undefine to save memory if semaphore is not needed
// #define TASK_USE_SEMAPHORE

// undefine to save memory if memory management is not needed
#define TASK_USE_MEM

// supported devices
#if defined (__AVR_ATmega168__)
#define RAMSTART 	0x60
#define HEAPSTART	0x02FF
#define STACK_SIZE	96
#define TIMER_REGISTER	TCCR0B
#define TIMER_ENABLE	TIMSK0
#define TIMER_VECTOR	TIMER0_OVF_vect
#define CLK1		_BV(CS00)
#define CLK1024		_BV(CS02) |_BV(CS00)
#endif

//some values

#define TRUE 1
#define FALSE 0
#define ERR 255

//number of tasks in the target system (incl. the idle task)

#define TASKS 4

//makros

#define TASK_ENTER_CRITICAL sreg=SREG;wdt_disable();cli()
#define TASK_LEAVE_CRITICAL sei();wdt_enable(WDTO_2S);SREG=sreg
#define MULTITASKING_DISABLE wdt_disable(); TIMER_ENABLE &= ~_BV(TOIE0)
#define MULTITASKING_ENABLE wdt_enable(WDTO_2S); TIMER_ENABLE |= _BV(TOIE0)
#define CONVERT_POINTER_TYPE_TO_MESSAGE(pointer) (void*)pointer
#define CONVERT_POINTER_TYPE_FROM_MESSAGE(message,type) (type)message.pointer


//+++++++++++++++++++++++++MAX TASKS++++++++++++++++++++
// max number of tasks - depending on ram amount
#define MAXTASKS ((RAMEND-RAMSTART-100)/(STACK_SIZE+2))

//+++++++++++++++++++TASK Function Dummy++++++++++++++++
typedef void (*Function)(void);

//+++++++++++++++++++TASK Priorities++++++++++++++++++++
// times & cycles at CLK/1024:
typedef enum Priority
{
	Low	= 0xff,	// 1024 cycles
	Medium	= 0xc0,	// 64*1024 cycles
	High	= 0x00	// 256*1024 cycles
} Priority;

//++++++++++++++++++TASK State++++++++++++++++++++++++++
#ifdef TASK_USE_SLEEP
typedef enum State
{
	Active = 0, 
	Sleeping = 1,
	SleepForLaps = 2,
} State;
#endif

//+++++++++++++++++TASK Control Structure+++++++++++++++
typedef volatile struct Task
{
	uint8_t ID;
	Priority priority;
#ifdef TASK_USE_SLEEP
	State state;
	uint8_t laps;
#endif
	unsigned char sreg;
	unsigned char *sp;
	unsigned char *stack_section;
} Task;

//+++++++++++++++++++++++SEMAPHORE+++++++++++++++++++++

#ifdef TASK_USE_SEMAPHORE

typedef volatile struct Semaphor
{
volatile uint8_t owner;
} Semaphor;

#endif

//+++++++++++++++++++++MESSAGE QUEUE+++++++++++++++++++

#ifdef TASK_USE_MESSAGE
#define MAXQUEUEDEPTH 2

typedef volatile struct Queue
{
	uint8_t item_size;
	uint8_t queueDepth;
	uint8_t number;
	void* ptr;
	unsigned char *message;
} Queue;

#endif

//+++++++++++++++++MEMORY MANAGEMENT+++++++++++++++++++
#ifdef TASK_USE_MEM
#define MAXBLOCKS 2
#define BLOCKSIZE 2

typedef volatile struct mem_block_intern
{
	uint8_t owner;
	volatile void* ptr;
} mem_block_intern;

static volatile char data[MAXBLOCKS*BLOCKSIZE];
static volatile mem_block_intern blocks[MAXBLOCKS];

#endif

//+++++++++++++++++++++++PROTOTYPE+++++++++++++++++++++

static void kernel(void);
void TIMER_VECTOR(void) __attribute__ ( ( signal, naked ) ); //see scheduler.c 
void idle(void);

static void init_user_environment(void);
static void shutdown_user_environment(void);

void reset_system(void);
void shutdown_system(void);

#ifdef TASK_USE_MESSAGE //see message.c
uint8_t queue_init(Queue *qp, uint8_t MessageQueueDepth, uint8_t item_size);
uint8_t queue_remove(Queue *qp);
uint8_t queue_exists(Queue *qp); 
uint8_t queue_get_maximum_size(Queue *qp); 
uint8_t queue_get_free_size(Queue *qp);
uint8_t queue_get_number_of_messages(Queue *qp);
uint8_t queue_wait_for_message(Queue *qp);
unsigned char queue_read_byte(Queue *qp, uint8_t position, uint8_t delete_message);
uint8_t queue_read_item(Queue *qp, unsigned char* targetpointer, uint8_t position, uint8_t delete_message);
void* queue_read_pointer (Queue *qp);
uint8_t queue_write_byte (Queue *qp, unsigned char message);
uint8_t queue_write_item(Queue *qp, void* startpointer);
uint8_t queue_write_pointer (Queue *qp, void* pointer);
#endif

#ifdef TASK_USE_SEMAPHORE //see semaphore.c 
void semaphore_init(Semaphor *s);
uint8_t semaphore_set(Semaphor *s);
uint8_t semaphore_free(Semaphor *s);
uint8_t semaphore_read(Semaphor *s);
uint8_t semaphore_wait(Semaphor *s);
#endif

#ifdef TASK_USE_MEM //see memory.c
uint8_t memory_get_free_blocks(void);
void* memory_allocate_block(void);
uint8_t memory_free_block(void* ptr);
#endif

//see task_control.c
uint8_t task_insert(Task *t, Function func_name, Priority prio);
void task_schedule(void);
uint8_t task_new_prio(Priority prio);

#ifdef TASK_USE_ANHILATE //see task_control.c
uint8_t task_remove(Task *t);
#endif

#ifdef TASK_USE_SLEEP //see task_control.c
uint8_t task_wake(Task *t);
uint8_t task_sleep(Task *t);
uint8_t task_sleep_for_laps(Task *t, uint8_t laps);
#endif

//++++++++++++++++++++GlOBAL VARIABLES+++++++++++++++++

static volatile signed char active_task_id = -1;
#ifdef TASK_USE_MEM
static volatile uint8_t activeBlocks = 0;
#endif
static volatile Task *task_container[TASKS], *active_task = 0;
static volatile uint8_t sreg = 0, sregsys = 0;
volatile unsigned char stack_global[TASKS*STACK_SIZE];
Task idletask;

